import orjson
import numpy as np
from datetime import datetime

"""
A custom class for making data objects that are easily exported/imported as json. Uses orjson for fast export and 
import of numpy arrays. 

The export and import methods can handle DataObj objects inside other DataObj objects. 

Why not just use the pickle library? 
--> Answer: Json is a more universal format than python pickle. The exports from this library could be parsed by Matlab, 
Javascript, etc. The use of orjson here should also make this much faster than python pickle for large datasets. 
"""


class DataObj:
    def __init__(self, name=None, dic=None):
        if name is not None:
            self.load_file(name)
        if dic is not None:
            self.load_dic(dic)

    def export(
        self, name, include_time=False, print_info=False, include_time_inside=False
    ):
        dic = self.export_dic()
        if include_time:
            now = datetime.now()
            dt_string = now.strftime("%d.%m.%Y_%H.%M.%S")
            name = name + dt_string
        if include_time_inside:
            now = datetime.now()
            dt_string = now.strftime("%d.%m.%Y_%H.%M.%S")
            dic["date_time"] = dt_string
        strb = orjson.dumps(dic, option=orjson.OPT_SERIALIZE_NUMPY)

        if name[-5:] != ".json":
            json_name = name + ".json"
        else:
            json_name = name
        with open(json_name, "wb") as file:
            file.write(strb)
        if print_info:
            print("Saving data as: ", name)
        return name  # if I want to save other file types with same name

    def load_file(self, name):
        with open(name, "rb") as file:
            strb = file.read()
        self.load_dic(orjson.loads(strb))
        print(self.__dict__.keys())


    # recursive
    def check_list(self, item):
        if type(item) is list:
            return self.check_list(item[0])
        if (type(item) is float) or (type(item) is int):
            return True
        else:
            return False

    def load_dic(self, dic):

        dic_2 = dic.copy()
        for key in dic.keys():
            # handle numpy arrays
            # does not working with multi-dimensional numpy arrays yet. 
            # need recursive check function so that lists of lists of numpy becomes np.array(np.array(np.array))
            if self.check_list(dic[key]):
                dic_2[key] = np.array(dic[key])

            # handle lists of objects of type JsonTool
            if (type(dic[key]) is list) and (key[-3:] == "_do"):
                ls = dic_2[key]
                del dic_2[key]
                for i in range(len(ls)):
                    ls[i] = DataObj(dic=ls[i])
                dic_2[key[:-3]] = ls
                continue

            # handle keys with value of type JsonTool
            if key[-3:] == "_do":
                sub_dic = dic[key]
                del dic_2[key]
                dic_2[key[:-3]] = DataObj(dic=sub_dic)
        self.__dict__.update(dic_2)

    def export_dic(self):
        dic = self.__dict__
        dic_2 = dic.copy()
        for key in dic.keys():
            # print(type(dic[key]))

            # print(issubclass(dic[key].__class__, type(self)))
            # print(issubclass(type(dic[key]), type(self)))
            # print("###########")
            # print(type(self))
            # print(type(dic[key]))
            # print(issubclass(type(dic[key]), type(self)))
            #
            # print("###########")
            # print(dic[key].__bases__)
            if (type(dic[key]) is type(self)) or issubclass(type(dic[key]), type(self)):
                # dic[key] = dic[key].export_dic()
                json_tool_dic = dic[key].export_dic()
                del dic_2[key]
                new_key = str(key) + "_do"
                dic_2[new_key] = json_tool_dic

            if (type(dic[key]) is list) and (
                (type(dic[key][0]) is type(self))
                or issubclass(type(dic[key][0]), type(self))
            ):
                ls = []
                for i, item in enumerate(dic[key]):
                    ls.append(dic[key][i].export_dic())

                del dic_2[key]
                new_key = str(key) + "_do"
                dic_2[new_key] = ls
        return dic_2


# would be cool if this could overload the list functionality if I only wanted to use this object as a list.
# right now it does not encode lists of numpy arrays correctly.


if __name__ == "__main__":
    # create a DataObj with various things inside, including a list of DataObj, a single DataObj,
    # and an extra numpy array.
    parent = DataObj()
    parent.items = []
    # for i in range(20):
    #     struct = DataObj()
    #     struct.item = "this is an item"
    #     struct.dB = f"{i} dB"
    #     struct.ls = [
    #         np.array([3, 34, 2, 43]),
    #         np.array([3, 34, 2, 43]),
    #         np.array([3, 34, 2, 43]),
    #     ]
    #     struct.arr = np.array([23, 2, 34, 2, 43, 3, 4.433])
    #     parent.items.append(struct)
    # parent.extraObj = DataObj()
    # parent.extraObj.item = "this is an item inside an nested DataObj"
    # parent.numpy = np.array([3, 54, 3, 45, 2, 5, 7, 3])
    # # export
    # parent.export("try_this.json")

    # # import
    # parent_2 = DataObj("try_this.json")
    # print(type(parent_2))
    # for item in parent_2.__dict__.keys():
    #     print(item)
    # print(parent_2.extraObj.item)
    # print()
    # print(parent_2.numpy)
    # print(type(parent_2.numpy))
    # print()
    # print([parent_2.items[i].dB for i in range(10)])
    # print([parent_2.items[i].arr for i in range(10)])
    parent.items.append([np.array([4,5,23,45,3,4]),np.array([4,5,23,45,3,4]),np.array([55,5,23,45,3,4])])
    parent.export("try_this.json")
    print()
    parent_2 = DataObj("try_this.json")
    print(parent_2.items)
    print(type(parent_2.items))
    # print(parent.check_numpy(parent.items))






import numpy as np
from numba import njit
import numba
import math


@njit
def clockLock(
    _channels,
    _timetags,
    clockChan,
    dataChan,
    pulses_per_clock,
    phase,
    window=0.5,
    deriv=1800,
    prop=2e-12,
    guardPeriod=0,
):
    """
    This clock locking function takes in clock timetags on the clockChan variable, and data on the dataChan variable.
    It distributes an accurate virtual clock to muliple times bewteen each clock using a PLL.
    :param _channels:
    :param _timetags:
    :param clockChan:
    :param dataChan:
    :param pulses_per_clock: number of snspd/data tags per raw clock tag.
    :param phase:
    :param window: an area of time around each distributed clock time (pulse time) that
    is seached for timetags. Tags outside that range are discarded (this is useful
    for scanning for the phase offset of clock and snspd tags)
    :param deriv: the derivative parameter for the PLL
    :param prop: the proportional parameter for the PLL
    :param guardPeriod: number of actual clock tags that are used to let the PLL stabilize, but are not used for
    outputting data because the locked clock can oscillate a little before it stabilizes.
    :return:
    """
    j = 0
    k = 0
    phi0 = 0
    phiold = 0
    filterr = 0
    clock0 = -1
    cd = 0
    holdup = False  # a flag that is usually off, used for printing certain errors
    Clocks = np.zeros(len(_channels))
    dataTags = np.zeros(len(_channels))
    ClockPortion = np.zeros(1000)

    for i in range(1000):
        if _channels[i] == clockChan:
            ClockPortion[j] = _timetags[i]
            j = j + 1
    j = 0

    # Initial Estimates
    ClockPortion = ClockPortion[ClockPortion > 0]  # cut off extra zeros
    period = (ClockPortion[-1] - ClockPortion[0]) / (len(ClockPortion) - 1)
    freq = 1 / period

    RecoveredClocks = np.zeros(len(_channels))
    Periods = np.zeros(len(_channels))
    nearestPulseTimes = np.zeros(len(_channels))
    Cycles = np.zeros(len(_channels))

    for i in range(len(_channels)):
        if _channels[i] == clockChan:
            currentClock = _timetags[i]
            if clock0 == -1:
                clock0 = currentClock - period

            arg = ((currentClock - (clock0 + period)) / period) * 2 * math.pi
            phi0 = math.sin(arg)
            filterr = phi0 + (phi0 - phiold) * deriv
            freq = freq - filterr * prop
            clock0 = clock0 + (1 / freq)  # add one period
            period = 1 / freq
            phiold = phi0
            j = j + 1
            if j >= guardPeriod:
                RecoveredClocks[j - guardPeriod] = clock0
                Periods[j - guardPeriod] = period
                Clocks[j - guardPeriod] = _timetags[i]

        if _channels[i] == dataChan:
            if clock0 != -1 and j >= guardPeriod:
                tag = _timetags[i]
                hist_tag = _timetags[i] - clock0
                binTime = period / pulses_per_clock  # about 2000ps for now
                # dist = period/binTime # 64 for this awg run
                cycles = (
                    hist_tag + phase
                ) / binTime  # cycles ranges from a to a + 64 in this case
                if abs(cycles - round(cycles)) <= window:

                    # only add the data if its near a laser pulse
                    Cycles[k] = cycles
                    cycles = round(cycles)
                    nearestPulseTimes[k] = cycles * binTime + clock0
                    dataTags[k] = tag
                    k = k + 1
            else:
                # no usable recovered clock available yet. Throw out that data
                continue

    Clocks = Clocks[Clocks != 0]
    RecoveredClocks = RecoveredClocks[RecoveredClocks != 0]
    dataTags = dataTags[dataTags != 0]
    Periods = Periods[Periods != 0]
    nearestPulseTimes = nearestPulseTimes[nearestPulseTimes != 0]
    return Clocks, RecoveredClocks, dataTags, nearestPulseTimes, Cycles


@njit
def agnostic_clock_lock(
    _channels,
    _timetags,
    clockChan,
    dataChan,
    pulses_per_clock,
    phase,
    window=0.5,
    deriv=1800,
    prop=2e-12,
    guardPeriod=0,
    clock_mult=200,
):
    j = 0
    k = 0
    phi0 = 0
    phiold = 0
    filterr = 0
    clock0 = -1
    cd = 0
    holdup = False  # a flag that is usually off, used for printing certain errors
    Clocks = np.zeros(len(_channels))
    dataTags = np.zeros(len(_channels))
    ClockPortion = np.zeros(1000)

    for i in range(1000):
        if _channels[i] == clockChan:
            ClockPortion[j] = _timetags[i]
            j = j + 1
    j = 0

    # Initial Estimates
    ClockPortion = ClockPortion[ClockPortion > 0]  # cut off extra zeros
    period = (ClockPortion[-1] - ClockPortion[0]) / (len(ClockPortion) - 1)
    freq = 1 / period

    RecoveredClocks = np.zeros(len(_channels))
    Periods = np.zeros(len(_channels))
    nearestPulseTimes = np.zeros(len(_channels))
    Cycles = np.zeros(len(_channels))
    rel_clocks = np.zeros(len(_channels))

    for i in range(len(_channels)):
        if _channels[i] == clockChan:
            currentClock = _timetags[i]
            if clock0 == -1:
                clock0 = currentClock - period

            arg = ((currentClock - (clock0 + period)) / period) * 2 * math.pi
            phi0 = math.sin(arg)
            filterr = phi0 + (phi0 - phiold) * deriv
            freq = freq - filterr * prop
            clock0 = clock0 + (1 / freq)  # add one period
            period = 1 / freq
            phiold = phi0
            j = j + 1
            if j >= guardPeriod:
                RecoveredClocks[j - guardPeriod] = clock0
                Periods[j - guardPeriod] = period
                Clocks[j - guardPeriod] = _timetags[i]

        if _channels[i] == dataChan:
            if clock0 != -1 and j >= guardPeriod:
                tag = _timetags[i]
                hist_tag = _timetags[i] - clock0

                sub_period = (
                    period / clock_mult
                )  # this is dumb. It's exactly the same as period/pulses_per_clock, right?
                sub_addition = hist_tag // sub_period

                binTime = period / pulses_per_clock  # about 2000ps for now
                # dist = period/binTime # 64 for this awg run
                cycles = (
                    hist_tag + phase
                ) / binTime  # cycles ranges from a to a + 64 in this case

                # only add the data if its near a laser pulse
                Cycles[k] = cycles
                cycles = round(cycles)
                nearestPulseTimes[k] = cycles * binTime + clock0
                dataTags[k] = tag
                rel_clocks[k] = clock0 + sub_addition * sub_period
                k = k + 1
            else:
                # no usable recovered clock available yet. Throw out that data
                continue

    Clocks = Clocks[Clocks != 0]
    RecoveredClocks = RecoveredClocks[RecoveredClocks != 0]
    dataTags = dataTags[dataTags != 0]
    Periods = Periods[Periods != 0]
    nearestPulseTimes = nearestPulseTimes[nearestPulseTimes != 0]
    rel_clocks = rel_clocks[rel_clocks != 0]
    return Clocks, RecoveredClocks, dataTags, nearestPulseTimes, Cycles, rel_clocks


# used for phase locked loop in real time Swabian software. Layout based on an example for a CustomMeasurment
# from the Swabian documentation.
class RepeatedPll:
    def __init__(
        self,
        tags_per_cycle,
        channels,
        timetags,
        clock_chan,
        data_chan,
        mult,  # if set to pulses_per_clock, then a virtual clock is aligned with each pulse time
        phase,
        window=0.5,
        deriv=1800,
        prop=2e-12,
        guard_period=0,
    ):

        self.tags_per_cycle = tags_per_cycle
        self.channels = channels
        self.timetags = timetags
        self.clock_chan = clock_chan
        self.data_chan = data_chan
        self.mult = mult
        self.phase = phase
        self.window = window
        self.deriv = deriv
        self.prop = prop
        self.guard_period = guard_period
        self.clock0 = 0
        self.period = 1
        self.phi_old = 0
        self.init = (
            1  # used during 1st run of jit-function to get a rough clock estimate
        )
        # NOTE this won't work if clocks might be missing.

        # print("length of channels: ", len(channels))
        self.hist_tags = np.zeros(len(channels))
        self.clocks = np.zeros(len(channels))
        self.recovered_clocks = np.zeros(len(channels))
        self.nearest_pulses = np.zeros(len(channels))

        # this is only to keep pycharm happy...
        self.hist_buffer = np.zeros(self.tags_per_cycle)
        self.clocks_buffer = np.zeros(self.tags_per_cycle)
        self.recovered_clocks_buffer = np.zeros(self.tags_per_cycle)
        self.nearest_pulses_buffer = np.zeros(self.tags_per_cycle)

    def process(self):
        cycles = (len(self.timetags) // self.tags_per_cycle) + 1
        starts = np.arange(0, len(self.timetags), self.tags_per_cycle)
        ends = np.roll(starts, -1)
        ends[-1] = len(self.timetags)
        hist_idx = 0
        clock_idx = 0

        for start, end in zip(starts, ends):
            # in the swabian version I won't do this extracting out. There will just be the buffers,
            # print(start)
            # input
            self.tags_buffer = self.timetags[start:end]
            self.channels_buffer = self.channels[start:end]

            # output
            # self.hist_buffer = self.hist_tags[start:end] # this is just a view, not a new array I think
            self.hist_buffer = np.zeros(self.tags_per_cycle)
            self.clocks_buffer = np.zeros(self.tags_per_cycle)
            self.recovered_clocks_buffer = np.zeros(self.tags_per_cycle)
            self.nearest_pulses_buffer = np.zeros(self.tags_per_cycle)

            (
                self.clock0,
                self.period,
                self.phi_old,
                self.init,
            ) = RepeatedPll.fast_process(
                self.tags_buffer,
                self.channels_buffer,
                self.hist_buffer,
                self.clocks_buffer,  #
                self.recovered_clocks_buffer,  #
                self.nearest_pulses_buffer,  #
                self.clock_chan,
                self.data_chan,
                self.init,
                self.clock0,
                self.period,
                self.phi_old,
                self.deriv,
                self.prop,
                self.guard_period,
                self.phase,
                self.mult,
            )

            # in the realtime code for the swabian, you won't keep a giant array that corresponds to self.hist_tags.
            # That's why here I'm using these lines to transfer the buffers into the large arrays.

            # this wont work they need to be variable length additions

            self.clocks[
                clock_idx : clock_idx + len(self.clocks_buffer)
            ] = self.clocks_buffer
            self.recovered_clocks[
                clock_idx : clock_idx + len(self.recovered_clocks_buffer)
            ] = self.recovered_clocks_buffer
            self.hist_tags[
                hist_idx : hist_idx + len(self.hist_buffer)
            ] = self.hist_buffer
            self.nearest_pulses[
                hist_idx : hist_idx + len(self.hist_buffer)
            ] = self.nearest_pulses_buffer

            hist_idx = hist_idx + len(self.hist_buffer)
            clock_idx = clock_idx + len(self.clocks_buffer)

        return self.clocks, self.recovered_clocks, self.hist_tags, self.nearest_pulses
        #  need: Clocks, RecoveredClocks, dataTags, nearestPulseTimes,

    @staticmethod
    @numba.jit(nopython=True, nogil=True)
    def fast_process(
        tags,
        channels,
        hist_buffer,
        clocks_buffer,
        recovered_clocks_buffer,
        nearest_pulses_buffer,
        clock_chan,
        data_chan,
        init,
        clock0,
        period,
        phi_old,
        deriv,
        prop,
        guard_period,
        phase,
        mult,
    ):
        # is the numpy array cleared after each call? That's probably the most reasonable.
        # might be more efficient if I initialize it in process, and then just clear it and re-use it. I know what the
        # max number of tags per fast_process cycle is.

        freq = 1 / period

        if init:
            # need a rough estimate of the period and frequency to start.
            j = 0
            clock_portion = np.zeros(1000)
            for i in range(1000):
                if channels[i] == clock_chan:
                    clock_portion[j] = tags[i]
                    j = j + 1
            j = 0

            # Initial Estimates
            clock_portion = clock_portion[clock_portion > 0]  # cut off extra zeros
            period = (clock_portion[-1] - clock_portion[0]) / (len(clock_portion) - 1)
            freq = 1 / period
            init = 0
            clock0 = -1

        j = 0  # I don't think I even need j...
        k = 0
        for tag, channel in zip(tags, channels):
            if channel == clock_chan:
                currentClock = tag
                clocks_buffer[j] = currentClock
                if clock0 == -1:
                    clock0 = currentClock - period

                arg = ((currentClock - (clock0 + period)) / period) * 2 * math.pi
                phi0 = math.sin(arg)
                filterr = phi0 + (phi0 - phi_old) * deriv
                freq = freq - filterr * prop
                clock0 = clock0 + (1 / freq)  # add one period
                recovered_clocks_buffer[j] = clock0
                period = 1 / freq
                phi_old = phi0
                j = j + 1

            if channel == data_chan:
                if clock0 != -1 and j >= guard_period:

                    hist_buffer[k] = tag
                    hist_tag = tag - clock0  # IMPORTATNT!
                    # hist_tag, sub_period = RepeatedPll.wrap_hist(hist_tag, mult, period)
                    sub_period = period / mult

                    # with these lines, cycles (below) is always going to be about 1
                    # minor_cycles = hist_tag // sub_period            # for mult method
                    # hist_tag = hist_tag - sub_period * minor_cycles  # for mult method

                    cycles = (hist_tag + phase) / sub_period  # for central method
                    cycles = round(cycles)  # for central method
                    nearest_pulses_buffer[k] = (
                        cycles * sub_period + clock0
                    )  # for central method
                    k = k + 1
                else:
                    # no usable recovered clock available yet. Throw out that data
                    continue

        clocks_buffer = clocks_buffer[clocks_buffer != 0]
        recovered_clocks_buffer = recovered_clocks_buffer[recovered_clocks_buffer != 0]
        hist_buffer = hist_buffer[hist_buffer != 0]
        nearest_pulses_buffer = nearest_pulses_buffer[nearest_pulses_buffer != 0]
        return (
            clock0,
            period,
            phi_old,
            init,
        )  # hist_buffer, clocks_buffer, recovered_clocks_buffer, nearest_pulses_buffer



# -*- coding: utf-8 -*-

"""Top level package for my PhD code"""


from . import viz
from . import obj
from . import hist
from . import help
from . import layout

# from . import clock


__author__ = """Andrew Mueller"""
__email__ = """andrewstermueller@gmail.com"""
__version__ = "0.0.1"



import numpy as np
from scipy import special
from scipy.interpolate import UnivariateSpline
from scipy.stats import rv_continuous
from collections import namedtuple

from scipy.optimize import curve_fit
import math


# in the SNSPD ccommunity, we are often interested in gaussian-like response functions,
# and their width metrics like FWHM.

# these tools help with fitting of these response functions, and finding width metrics.


class SplineTool:
    def __init__(self, hist, _bins):
        if len(_bins) > len(hist):
            _bins = _bins[:-1]
        assert len(hist) == len(_bins)
        self.hist = np.array(hist)
        self.bins = np.array(_bins)
        self.y_scale = np.max(hist)
        self.norm_hist = hist / self.y_scale
        self.x_scale = self.bins[-1] - self.bins[0]
        self.norm_bins = self.bins / self.x_scale
        self.smoothing = 0
        self.spline = UnivariateSpline(self.norm_bins, self.norm_hist, s=self.smoothing)

    def plot_spline(self, plot_bins, smoothing):
        self.smoothing = smoothing
        self.spline.set_smoothing_factor(self.smoothing)
        norm_plot_bins = plot_bins / self.x_scale
        norm_plot_points = self.spline(norm_plot_bins)
        return plot_bins, norm_plot_points * self.y_scale

    def full_width_at_level(self, level, smoothing=None):
        assert level < 1
        if smoothing is not None:
            self.smoothing = smoothing

        # norm_hist goes from 0 to 1, but the max point (1) may not be a good estimate
        # for the max used for FWHM. Use the max of the spline instead.
        shifted_level = self.spline_max().y * level

        spline = UnivariateSpline(
            self.norm_bins, self.norm_hist - shifted_level, s=self.smoothing
        )
        roots = spline.roots()
        # print("roots before: ", roots)
        roots = roots * self.x_scale
        # print("xscale: ", self.x_scale)

        if len(roots) < 2:
            raise ValueError(f"There more or less than 2 roots: {roots}")
            # print("There more or less than 2 roots")
            # print("The roots are: ", roots)
            return 1
        if len(roots) > 2:
            print(f"There are more than 2 roots. They are: {roots}")
            root_1 = roots[np.argmax(np.diff(roots))]
            root_2 = roots[np.argmax(np.diff(roots)) + 1]
        if len(roots) == 2:
            root_1 = roots[0]
            root_2 = roots[1]

        height = self.y_scale * shifted_level
        RootData = namedtuple("RootData", "width left right height level")
        return RootData(
            width=root_2 - root_1, left=root_1, right=root_2, height=height, level=level
        )

    @staticmethod
    def quadratic_spline_roots(spl):
        # from: https://stackoverflow.com/questions/50371298/find-maximum-minimum-of-a-1d-interpolated-function
        roots = []
        knots = spl.get_knots()
        for a, b in zip(knots[:-1], knots[1:]):
            u, v, w = spl(a), spl((a + b) / 2), spl(b)
            t = np.roots([u + w - 2 * v, w - u, 2 * v])
            t = t[np.isreal(t) & (np.abs(t) <= 1)]
            roots.extend(t * (b - a) / 2 + (b + a) / 2)
        return np.array(roots)

    def spline_max(self):
        cr_pts = self.quadratic_spline_roots(self.spline.derivative())
        cr_pts = np.append(
            cr_pts, (self.norm_bins[0], self.norm_bins[-1])
        )  # also check the endpoints of the interval
        cr_vals = self.spline(cr_pts)
        min_index = np.argmin(cr_vals)
        max_index = np.argmax(cr_vals)
        MaxData = namedtuple("MaxData", "x y")
        return MaxData(x=cr_pts[max_index] * self.x_scale, y=cr_vals[max_index])

    def plot_width_at_level_arrows(
        self, plot, level, width_label=True, color="black", lw=1.5, alpha=1, ls="-"
    ):
        root_data = self.full_width_at_level(level)
        plot.annotate(
            text="",
            xy=(root_data.left, root_data.height),
            xytext=(root_data.right, root_data.height),
            arrowprops=dict(
                color=color,
                alpha=alpha,
                lw=lw,
                ls=ls,
                shrinkA=0,
                shrinkB=0,
                patchA=None,
                patchB=None,
                arrowstyle="<->",
            ),
            label=f"full width at {level} max",
            bbox=dict(pad=0),
        )
        center = (root_data.left + root_data.right) / 2
        if width_label:
            plot.annotate(
                f"{round(root_data.width, 2)}",
                xy=(center, root_data.height),
                color=color,
                alpha=alpha,
                xytext=(0, -1),
                fontsize=8,
                textcoords="offset points",
                horizontalalignment="center",
                verticalalignment="top",
            )

        # useful
        # https://stackoverflow.com/questions/23344891/matplotlib-set-pad-between-arrow-and-text-in-annotate-function
        return plot

    def sigma(self, smoothing=None):
        fwhm = self.full_width_at_level(0.5, smoothing).width
        return fwhm / 2.355

    def fwhm(self, smoothing=None):
        return self.full_width_at_level(0.5, smoothing).width

    def gaussian_fit(self):
        # curve_fit expects something more or less centered
        x_center = self.spline_max().x
        fwhm_width = self.full_width_at_level(0.5, 0.0001).width
        left_bound = x_center - fwhm_width * 3
        right_bound = x_center + fwhm_width * 3
        left_idx = np.searchsorted(self.bins, left_bound)
        right_idx = np.searchsorted(self.bins, right_bound)

        x_bias = np.average(self.bins[left_idx:right_idx])
        x_portion = self.bins[left_idx:right_idx] - x_bias
        y_portion = self.hist[left_idx:right_idx]
        popt, pcov = curve_fit(gaussian, x_portion, y_portion)
        popt[1] = popt[1] + x_bias

        return popt, pcov

    def plot_gaussian_fit(self):
        popt, pcov = self.gaussian_fit()
        return self.bins, gaussian(self.bins, *popt)


def gaussian(x, amplitude, mean, stddev):
    return amplitude * np.exp(-(((x - mean) / stddev) ** 2))


# def find_nearest(array,value):
#     idx = np.searchsorted(array, value, side="left")
#     if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
#         return array[idx-1]
#     else:
#         return array[idx]


def gaussian_background(x, sigma, mu, back, l, r):
    "d was found by symbolically integrating in mathematica"
    n = back + (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(
        -0.5 * (((x - mu) / sigma) ** 2)
    )
    d = 0.5 * (
        2 * back * (-l + r)
        + special.erf((-l + mu) / (np.sqrt(2) * sigma))
        - special.erf((mu - r) / (np.sqrt(2) * sigma))
    )
    return n / d


class GaussianTool:
    class gaussian_bg(rv_continuous):
        "Gaussian distributionwithj Background parameter 'back'"

        def _pdf(self, x, sigma, mu, back):
            return gaussian_background(x, sigma, mu, back, self.a, self.b)



import inspect
import re

# from: https://stackoverflow.com/questions/32000934/python-print-a-variables-name-and-value
def prinfo(*args):
    frame = inspect.currentframe().f_back
    s = inspect.getframeinfo(frame).code_context[0]
    r = re.search(r"\((.*)\)", s).group(1)
    vnames = r.split(", ")
    for i,(var,val) in enumerate(zip(vnames, args)):
        print(f"{var} = {val}")
        



from typing import List
import matplotlib.pyplot as plt

"""For layout of matplotlib plots
"""


def bisect(
    layout: List[float],
    direction="vert",
    offset=0.5,
    spacing=0.05,
    absolute_spacing=False,
):
    x = layout[0]
    y = layout[1]
    width = layout[2]
    height = layout[3]

    if direction != "vert":
        # horizontal bisect
        if not absolute_spacing:
            spacing = spacing * height
        return [x, y, width, height * offset - spacing / 2], [
            x,
            height * offset + (spacing / 2),
            width,
            height * (1 - offset) - (spacing / 2),
        ]
    else:
        # vertical bisect
        if not absolute_spacing:
            spacing = spacing * width
        return [x, y, width * offset - spacing / 2, height], [
            width * offset + (spacing / 2),
            y,
            width * (1 - offset) - spacing / 2,
            height,
        ]


def margin(layout: List[float], w=0.1):
    x = layout[0]
    y = layout[1]
    width = layout[2]
    height = layout[3]
    return [x + w, y + w, width - w, height - w]


if __name__ == "__main__":

    def layout_basic():
        fig = plt.figure(figsize=(total_size_x, total_size_y))

        master_left, master_right = bisect(
            [0, 0, 1, 1],
            direction="vert",
            offset=0.5,
            spacing=divider_width,
            absolute_spacing=True,
        )

        left, bulk = bisect(master_left, direction="vert", offset=0.15, spacing=0.07)
        _, left = bisect(left, direction="horiz", offset=0.15, spacing=0.07)
        bottom, image = bisect(bulk, direction="horiz", offset=0.15, spacing=0.07)

        bottom_right, top_right = bisect(
            master_right, direction="horiz", offset=0.6, spacing=0.14
        )

        ax_left_bar = fig.add_axes(left)
        ax_bottom_bar = fig.add_axes(bottom)
        ax_main = fig.add_axes(image, sharex=ax_bottom_bar, sharey=ax_left_bar)

        ax_diagonal = fig.add_axes(top_right)
        ax_hist = fig.add_axes(bottom_right)



# import seaborn as sns
import numpy as np

# import os
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.path import Path
from matplotlib.patches import BoxStyle
from matplotlib.offsetbox import AnchoredText
import matplotlib
from IPython import get_ipython

# import altair as alt

# import altair as alt


def despine(ax, offset=2):
    if (type(ax) != np.ndarray) & (type(ax) != list):
        ax = [ax]
    for a in ax:
        a.spines["bottom"].set_position(("outward", offset))
        a.spines["left"].set_position(("outward", offset))


def color_palette(dark=False) -> tuple[dict, list]:
    """
    Returns my preferred color scheme
    """
    colors = {
        "dark_purple": "#5F2E88",
        "purple": "#9d71c7",
        "light_purple": "#c08df0",
        "pale_purple": "#dfd6e5",
        "dark_orange": "#F38227",
        "orange": "#E39943",
        "light_orange": "#EEBA7F",
        "pale_orange": "#f2d4b6",
        "dark_blue": "#3F60AC",
        "blue": "#7292C7",
        "light_blue": "#A5B3CC",
        "pale_blue": "#dae4f1",
        "dark_red": "#9C372F",
        "red": "#C76A6A",
        "light_red": "#E39C9D",
        "pale_red": "#edcccc",
        "dark_green": "#395A34",
        "green": "#688A2F",
        "light_green": "#B3CD86",
        "pale_green": "#d8e2c3",
        "dark_brown": "#764f2a",
        "brown": "#c2996c",
        "light_brown": "#e1bb96",
        "pale_brown": "#efccaf",
        "black": "#444147",
        "grey": "#EFEFEF",
        "gray": "#EFEFEF",
        "light_grey": "#6D6F72",
        "light_gray": "#6D6F72",
    }

    if dark:
        colors = {
            "light_purple": "#c08df0",
            "light_orange": "#EEBA7F",
            "light_blue": "#A5B3CC",
            "light_red": "#E39C9D",
            "light_green": "#B3CD86",
            "light_brown": "#e1bb96",
            "pale_brown": "#efccaf",
            "black": "#444147",
            "grey": "#EFEFEF",
            "gray": "#EFEFEF",
            "light_grey": "#6D6F72",
            "light_gray": "#6D6F72",
        }
    palette = [
        v
        for k, v in colors.items()
        if (v not in ["grey", "gray", "dark_purple", "light_grey"])
        and ("pale" not in [v])
    ]
    return colors, palette


def bokeh_basic():
    # because I always forget bokeh syntax
    print("colors, swatches = viz.bokeh_theme()\n\
from bokeh.plotting import figure, show\n\
import numpy as np\n\
from bokeh.io import output_notebook\n\
output_notebook(hide_banner=True)\n\
p = figure(width=800, height=400, output_backend='webgl')\n\
p.line(x=np.arange(100), y=np.sin(np.arange(100)*.2), line_color=colors['black'], line_width=3, line_alpha=0.6)\n\
show(p)"
    )


def update_colors(dark=False):
    if not dark:
        plt.rcParams.update(
            {
                # "axes.facecolor": "#fafafa",
                "axes.facecolor": "#ffffff",
                "axes.edgecolor": "#333333",
                "axes.labelcolor": "#333333",
                "text.color": "#333333",
                "xtick.color": "#333333",
                "ytick.color": "#333333",
                "legend.edgecolor": "#333333",
                "figure.facecolor": "white",
                "grid.alpha": 0.3,
            }
        )
        colors, palette = color_palette()
    else:
        plt.rcParams.update(
            {
                "axes.facecolor": "#3d3f4f",
                # "axes.facecolor": "#ffffff",
                "axes.edgecolor": "#dbdbdb",
                "axes.labelcolor": "#dbdbdb",
                "text.color": "#dbdbdb",
                "xtick.color": "#dbdbdb",
                "ytick.color": "#dbdbdb",
                "legend.edgecolor": "#dbdbdb",
                "figure.facecolor": "#2E303E",
                "grid.alpha": 0.3,
            }
        )
        colors, palette = color_palette(dark=True)
    # sns.set_palette(palette)
    return "done"


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def setup_svg_and_autoreload():
    if is_notebook():
        ipython = get_ipython()
        ipython.run_line_magic("config", "InlineBackend.figure_formats = ['svg']")
        ipython.run_line_magic("load_ext", "autoreload")
        ipython.run_line_magic("autoreload", "2")


def phd_style(
    dark_mode: bool = False,
    jupyterStyle: bool = False,
    axese_width: int = 0,
    text: int = 0,
    tick_length: int = 0,
    data_width: int = 0,
    grid: bool = False,
    svg_mode=True,
):
    """
    Sets the plotting style to my preference
    """
    axese_width_j = 0
    text_j = 0
    tick_length_j = 0
    data_width_j = 0
    if jupyterStyle:
        axese_width_j = 0.6
        text_j = 5
        tick_length_j = 1
        data_width_j = 1.5

    items = []
    for item_org, item_j in zip(
        [axese_width, text, tick_length, data_width],
        [axese_width_j, text_j, tick_length_j, data_width_j],
    ):
        if item_org == 0:
            items.append(item_j)
        else:
            items.append(item_org)

    rc = {
        # "axes.facecolor": "#fafafa",
        "axes.facecolor": "#ffffff",
        "font.family": "sans-serif",
        "font.family": "DejaVu Sans",
        "font.style": "normal",
        "pdf.fonttype": 42,
        "patch.linewidth": 3,
        "axes.edgecolor": "#333333",
        "axes.labelcolor": "#333333",
        "axes.spines.right": False,
        "axes.spines.top": False,
        "axes.spines.left": True,
        "axes.spines.bottom": True,
        "axes.axisbelow": True,
        "axes.linewidth": 0.6 + items[0],
        "axes.titlesize": 9 + items[1],
        "text.color": "#333333",
        "axes.grid": grid,
        "lines.linewidth": 0.75 + items[3],
        "lines.dash_capstyle": "round",
        "patch.linewidth": 0.25,
        # "grid.linestyle": "-",
        # "grid.linewidth": 0.75,
        # "grid.color": "#ffffff",
        "axes.labelsize": 8 + items[1],
        "xtick.color": "#333333",
        "ytick.color": "#333333",
        "xtick.labelsize": 6 + items[1],
        "ytick.labelsize": 6 + items[1],
        "xtick.major.size": 3 + items[2],
        "ytick.major.size": 3 + items[2],
        "xtick.major.width": 0.6 + items[0],
        "ytick.major.width": 0.6 + items[0],
        "xtick.major.pad": 3,
        "ytick.major.pad": 3,
        "xtick.minor.size": 0,
        "ytick.minor.size": 0,
        "legend.fontsize": 7 + items[1],
        "legend.frameon": True,
        "legend.edgecolor": "#333333",
        "axes.xmargin": 0.03,
        "axes.ymargin": 0.03,
        "figure.facecolor": "white",
        "figure.dpi": 250,
        "errorbar.capsize": 1,
        "savefig.bbox": "tight",
        "grid.alpha": 0.3,
    }

    if dark_mode:
        rc = {
            # "axes.facecolor": "black",
            "axes.facecolor": "#3d3f4f",
            "axes.facecolor": "#2E303E",
            "font.family": "sans-serif",
            "font.family": "DejaVu Sans",
            "font.style": "normal",
            "pdf.fonttype": 42,
            "patch.linewidth": 3,
            "axes.edgecolor": "#dbdbdb",
            "axes.labelcolor": "#dbdbdb",
            "axes.spines.right": False,
            "axes.spines.top": False,
            "axes.spines.left": True,
            "axes.spines.bottom": True,
            "axes.axisbelow": True,
            "axes.linewidth": 0.6 + items[0],
            "axes.titlesize": 9 + items[1],
            "text.color": "#444147",
            "axes.grid": grid,
            "lines.linewidth": 0.75 + items[3],
            "lines.dash_capstyle": "round",
            "patch.linewidth": 0.25,
            # "grid.linestyle": "-",
            # "grid.linewidth": 0.75,
            # "grid.color": "#ffffff",
            "axes.labelsize": 8 + items[1],
            "xtick.color": "#dbdbdb",
            "ytick.color": "#dbdbdb",
            "xtick.labelsize": 6 + items[1],
            "ytick.labelsize": 6 + items[1],
            "xtick.major.size": 3 + items[2],
            "ytick.major.size": 3 + items[2],
            "xtick.major.width": 0.6 + items[0],
            "ytick.major.width": 0.6 + items[0],
            "xtick.major.pad": 3,
            "ytick.major.pad": 3,
            "xtick.minor.size": 0,
            "ytick.minor.size": 0,
            "legend.fontsize": 7 + items[1],
            "legend.frameon": True,
            "legend.edgecolor": "#dbdbdb",
            "axes.xmargin": 0.03,
            "axes.ymargin": 0.03,
            "figure.facecolor": "#2E303E",
            "figure.dpi": 250,
            "errorbar.capsize": 1,
            "savefig.bbox": "tight",
            "grid.alpha": 0.3,
        }

    if svg_mode:
        setup_svg_and_autoreload()

    plt.rc("text.latex", preamble=r"\usepackage{mathpazo}")
    matplotlib.style.use(rc)
    if dark_mode:
        colors, palette = color_palette(dark=True)
    else:
        colors, palette = color_palette()
    # sns.set_palette(palette)
    return colors, palette


# def load_bokeh():


def bokeh_theme() -> tuple[dict, list]:
    import bokeh

    tick_and_line_color = "#4f4f4f"
    lw = 1.7
    theme_json = {
        "attrs": {
            "figure": {"background_fill_color": "#f7f8fa", "outline_line_color": None},
            "Axis": {
                "axis_line_color": tick_and_line_color,
                "axis_line_width": 1.5,
                "axis_line_cap": "round",
                "major_tick_line_width": lw,
                "minor_tick_line_width": lw,
                "major_tick_line_cap": "round",
                "minor_tick_line_cap": "round",
                "major_tick_out": 7,
                "major_tick_in": 0,
                "major_tick_line_color": tick_and_line_color,
                "minor_tick_line_color": tick_and_line_color,
                "axis_label_text_font_size": "14pt",
                "major_label_text_font_size": "14pt",
                # "axis_label_text_font": "Open Sans Semibold",
                # "major_label_text_font": "Open Sans Semibold",
            },
            "legend_label": {
                "border_line_color": "black",
                "background_fill_color": "#EEEEEE",
                "border_line_width": 0.75,
                "background_fill_alpha": 0.75,
                "legend_font_size": "20pt",
            },
            "Grid": {
                "grid_line_alpha": 0.06,
                "grid_line_color": "black",
                "grid_line_width": 1.3,
                "grid_line_dash": "solid",
            },
            # "WheelZoomTool": {
            #     "dimension": 'width'
            # },
            "ToolbarBase": {
                "autohide": False,
                "logo": None,
            },
            "Text": {
                "text_font_style": "regular",
                "text_font_size": "48pt",
            },
            "Title": {
                "background_fill_color": "#FFFFFF",
                "text_color": "#3c3c3c",
                "align": "left",
                # "text_font": "Open Sans Semibold",
                "text_font_style": "normal",
                "text_font_size": "14pt",
                "offset": 0,
            },
            "Line": {
                "line_width": 2.2,
                # "line_color": '#b00c00'
            },
        }
    }

    colors, palette = color_palette()
    theme = bokeh.themes.Theme(json=theme_json)
    bokeh.themes
    bokeh.io.curdoc().theme = theme
    
    return (colors, palette)
    


# def altair_theme():
#     """
#     Sets a theme for the plotting library Altair to match the style of my PhD.
#     """
#
#     colors, palette = color_palette()
#     def _theme():
#         return {
#             'config': {
#                 'background': 'white',
#                     'group': {
#                     'fill': 'white',
#                     },
#                 'view': {
#                     'strokeWidth': 3,
#                     'height': 300,
#                     'width': 400,
#                     'fill': '#EEEEEE'
#                     },
#                 'mark': {
#                     'strokeWidth': 2,
#                     'stroke': 'black'
#                 },
#                 'axis': {
#                     'domainColor': colors['black'],
#                     'domainWidth': 0.5,
#                     'labelColor': colors['black'],
#                     'labelFontSize': 8,
#                     'labelFont': 'Myriad Pro',
#                     'titleFont': 'Myriad Pro',
#                     'titleFontWeight': 400,
#                     'titleFontSize': 8,
#                     'grid': True,
#                     'gridColor': 'white',
#                     'gridWidth': 0.75,
#                     'ticks': True,
#                     'tickColor': 'white',
#                     'tickOffset': 8,
#                     'tickWidth': 1.5
#                 },
#                 'Grid': {
#                     'grid_line_dash': [6,4]
#                 }
#                 'range': {
#                     'category': palette
#                 },
#                 'legend': {
#                     'labelFontSize': 8,
#                     'labelFont': 'Myriad Pro',
#                     'titleFont': 'Myriad Pro',
#                     'titleFontSize': 8,
#                     'titleFontWeight': 400
#                 },
#                 'title' : {
#                     'font': 'Myriad Pro',
#                     'fontWeight': 400,
#                     'fontSize': 8,
#                     'anchor': 'middle'
#                 }
#                   }
#                 }
#
#     alt.themes.register('phd', _theme)# enable the newly registered theme
#     alt.themes.enable('phd')
#     return colors, palette


def color_selector(style):
    """
    Select the color palette of your choice.

    Parameters
    ----------
    style: str "mut" or "pboc"
        A string identifier for the style. "mut" gives colors for single and double mutants.
        "pboc" returns the PBoC2e color palette.

    Returns
    -------
    colors: dict
        Dictionary of colors. If "dna", "double", or "inducer" is the selected style,
        keys will be the mutants in upper case. Double mutant keys will be DNA-IND. For
        pboc, the keys will be the typical color descriptors.

    """
    # Ensure the provided style name makes sense.
    if style.lower() not in ["mut", "pboc"]:
        raise ValueError(
            "Provided style must be 'pboc' or 'mut'. {} provided.".format(style)
        )

    # Set the color styles and return.
    if style.lower() == "mut":
        colors = {
            "Y20I": "#738FC1",
            "Q21A": "#7AA974",
            "Q21M": "#AB85AC",
            "F164T": "#A97C50",
            "Q294K": "#5D737E",
            "Q294V": "#D56C55",
            "Q294R": "#B2AF58",
            "Y20I-F164T": "#2d98da",
            "Y20I-Q294K": "#34495e",
            "Y20I-Q294V": "#8854d0",
            "Q21A-F164T": "#4b6584",
            "Q21A-Q294K": "#EE5A24",
            "Q21A-Q294V": "#009432",
            "Q21M-F164T": "#1289A7",
            "Q21M-Q294K": "#6F1E51",
            "Q21M-Q294V": "#006266",
            "WT": "#3C3C3C",
        }

    elif style.lower() == "pboc":
        colors = {
            "green": "#7AA974",
            "light_green": "#BFD598",
            "pale_green": "#DCECCB",
            "yellow": "#EAC264",
            "light_yellow": "#F3DAA9",
            "pale_yellow": "#FFEDCE",
            "blue": "#738FC1",
            "light_blue": "#A9BFE3",
            "pale_blue": "#C9D7EE",
            "red": "#D56C55",
            "light_red": "#E8B19D",
            "pale_red": "#F1D4C9",
            "purple": "#AB85AC",
            "light_purple": "#D4C2D9",
            "dark_green": "#7E9D90",
            "dark_brown": "#905426",
        }
    return colors


def titlebox(
    ax, text, color, bgcolor=None, size=8, boxsize="10%", pad=0.02, loc=10, **kwargs
):
    """Sets a colored box about the title with the width of the plot"""
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("top", size=boxsize, pad=pad)
    cax.get_xaxis().set_visible(False)
    cax.get_yaxis().set_visible(False)
    cax.spines["top"].set_visible(True)
    cax.spines["right"].set_visible(True)
    plt.setp(cax.spines.values(), color=color)
    if bgcolor != None:
        cax.set_facecolor(bgcolor)
    else:
        cax.set_facecolor("white")
    at = AnchoredText(text, loc=loc, frameon=False, prop=dict(size=size, color=color))
    cax.add_artist(at)


def ylabelbox(ax, text, color, bgcolor=None, size=6, boxsize="15%", pad=0.02, **kwargs):
    """Sets a colored box about the title with the width of the plot"""
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("left", size=boxsize, pad=pad)
    cax.get_xaxis().set_visible(False)
    cax.get_yaxis().set_visible(False)
    cax.spines["top"].set_visible(True)
    cax.spines["right"].set_visible(True)
    plt.setp(cax.spines.values(), color=color)
    if bgcolor != None:
        cax.set_facecolor(bgcolor)
    else:
        cax.set_facecolor("white")

    at = AnchoredText(
        text,
        loc=10,
        frameon=False,
        prop=dict(rotation="vertical", size=size, color=color),
    )
    cax.add_artist(at)
